#include <cmath>
#include "mex.h"
#include "quern.h"

// usage:
// [x,flag,relres,iter]=quern_cgnr_mex(A, b, tol, maxit, drop_tolerance)
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
   // check # arguments
   if(nrhs<2) mexErrMsgTxt("need at least two arguments");
   if(nlhs==0) return;
   if(nlhs>4) mexErrMsgTxt("at most four outputs");
   // get matrix A
   if(mxGetNumberOfDimensions(prhs[0])!=2)
      mexErrMsgTxt("Input matrix A is not two dimensional");
   if(!mxIsSparse(prhs[0]))
      mexErrMsgTxt("Input matrix A is not sparse");
   int m=mxGetM(prhs[0]), n=mxGetN(prhs[0]);
   const int* A_column_start=mxGetJc(prhs[0]);
   const int* A_row_index=mxGetIr(prhs[0]);
   const double* A_CSR_value=mxGetPr(prhs[0]);
   int nnz=A_column_start[n];
   int* A_row_start=(int*)mxMalloc((m+1)*sizeof(int));
   int* A_column_index=(int*)mxMalloc(nnz*sizeof(int));
   double* A_value=(double*)mxMalloc(nnz*sizeof(double));
   if(QUERN_convert_column_format_to_row_format(m, n,
                      A_column_start, A_row_index, A_CSR_value,
                      A_row_start, A_column_index, A_value)){
      mexErrMsgTxt("QUERN failure");
   }
   // get b
   if(mxGetN(prhs[1])!=1)
      mexErrMsgTxt("can only solve one vector at a time");
   if(mxGetM(prhs[1])!=m && mxGetM(prhs[1])!=n)
      mexErrMsgTxt("wrong size for right-hand side vector");
   const double* b=mxGetPr(prhs[1]);
   // get convergence tolerance
   double relative_tolerance=1e-6;
   if(nrhs>=3 && mxGetM(prhs[2])*mxGetN(prhs[2])>=1)
      relative_tolerance=*mxGetPr(prhs[2]);
   // get max its
   int maxit=100;
   if(nrhs>=4 && mxGetM(prhs[3])*mxGetN(prhs[3])>=1)
      maxit=(int)*mxGetPr(prhs[3]);
   // and drop tolerance
   double drop_tolerance=0.1;
   if(nrhs>=5 && mxGetM(prhs[4])*mxGetN(prhs[4])>=1)
      drop_tolerance=*mxGetPr(prhs[4]);

   // form A^T*b if this is least-squares
   double *rhs=0;
   if(mxGetM(prhs[1])==m){
      rhs=(double*)mxMalloc(n*sizeof(double));
      QUERN_multiply_transpose(m, n, A_row_start, A_column_index, A_value,
                               b, rhs);
   }

   // figure out absolute tolerance
   double initial_residual=0;
   if(rhs) for(int i=0; i<n; ++i) initial_residual+=rhs[i]*rhs[i];
   else for(int i=0; i<n; ++i) initial_residual+=b[i]*b[i];
   initial_residual=std::sqrt(initial_residual);
   double tol=relative_tolerance*initial_residual;

   // find the preconditioner
   // use profile row ordering
   int* row_order=(int*)mxMalloc(m*sizeof(int));
   if(QUERN_get_profile_row_ordering(m, n, A_row_start, A_column_index,
                                     row_order))
      mexErrMsgTxt("QUERN failure");
   // get R
   int* R_row_start=0;
   int* R_column_index=0;
   double* R_value=0;
   if(QUERN_compute_incomplete_qr_without_q(m, n, A_row_start,
                            A_column_index, A_value, row_order, drop_tolerance,
                            &R_row_start, &R_column_index, &R_value))
      mexErrMsgTxt("QUERN failure");

   // get output
   plhs[0]=mxCreateNumericMatrix(n, 1, mxDOUBLE_CLASS, mxREAL);
   double* x=mxGetPr(plhs[0]);

   // run CGNR
   int solved=-1, iterations=-1;
   double residual_norm=-1;
   if(QUERN_solve_with_CGNR(m, n, A_row_start, A_column_index, A_value,
                           rhs ? rhs : b, R_row_start, R_column_index, R_value,
                           maxit, tol, x, &solved, &iterations, &residual_norm))
      mexErrMsgTxt("QUERN failure");

   // output statistics if desired
   if(nlhs>1){ // flag
      plhs[1]=mxCreateNumericMatrix(1, 1, mxDOUBLE_CLASS, mxREAL);
      *mxGetPr(plhs[1])=!solved;
   }
   if(nlhs>2){ // relres
      plhs[2]=mxCreateNumericMatrix(1, 1, mxDOUBLE_CLASS, mxREAL);
      *mxGetPr(plhs[2])=residual_norm/initial_residual;
   }
   if(nlhs>3){ // iter
      plhs[3]=mxCreateNumericMatrix(1, 1, mxDOUBLE_CLASS, mxREAL);
      *mxGetPr(plhs[3])=(double)iterations;
   }

   QUERN_free_result(R_row_start, R_column_index, R_value);
   QUERN_free_result(A_row_start, A_column_index, A_value);
}

