#include <cstring>
#include "mex.h"
#include "quern.h"

// code to generally check anything in the Quern API using Matlab

mxArray* create_sparse_flipped(int m, int n, const int* row_start,
                               const int* column_index, const double* value)
{
   mxArray* matrix=mxCreateSparse(m, n, row_start[m], mxREAL);
   int* column_start=mxGetJc(matrix);
   int* row_index=mxGetIr(matrix);
   double* new_value=mxGetPr(matrix);
   if(QUERN_convert_column_format_to_row_format(n, m, row_start, column_index,
         value, column_start, row_index, new_value))
      mexErrMsgTxt("QUERN failure");
   return matrix;
}

mxArray* create_dense(int n, const int* value, int shift=1)
{
   mxArray* matrix=mxCreateNumericMatrix(n, 1, mxDOUBLE_CLASS, mxREAL);
   for(int i=0; i<n; ++i)
      mxGetPr(matrix)[i]=(double)(value[i]+shift);
   return matrix;
}

mxArray* create_dense(int n, const double* value)
{
   mxArray* matrix=mxCreateNumericMatrix(n, 1, mxDOUBLE_CLASS, mxREAL);
   for(int i=0; i<n; ++i)
      mxGetPr(matrix)[i]=value[i];
   return matrix;
}

// usage: quern_internal(name, A, options)
// where name is a string specifying the operation, A is a sparse matrix,
// and options is an array of numbers for the routine
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
   // check # arguments
   if(nrhs!=3) mexErrMsgTxt("need three arguments");
   if(nlhs==0) return;
   // check name of Quern function
   if(!mxIsChar(prhs[0])) mexErrMsgTxt("first argument must be a string");
   char funcname[20];
   mxGetString(prhs[0], funcname, 20);
   // check matrix A
   if(mxGetNumberOfDimensions(prhs[1])!=2)
      mexErrMsgTxt("Input matrix A is not two dimensional");
   if(!mxIsSparse(prhs[1]))
      mexErrMsgTxt("Input matrix A is not sparse");
   int m=mxGetM(prhs[1]), n=mxGetN(prhs[1]);
   const int* A_column_start=mxGetJc(prhs[1]);
   const int* A_row_index=mxGetIr(prhs[1]);
   const double* A_CSR_value=mxGetPr(prhs[1]);
   int nnz=A_column_start[n];
   int* A_row_start=(int*)mxMalloc((m+1)*sizeof(int));
   int* A_column_index=(int*)mxMalloc(nnz*sizeof(int));
   double* A_value=(double*)mxMalloc(nnz*sizeof(double));
   if(QUERN_convert_column_format_to_row_format(m, n,
                      A_column_start, A_row_index, A_CSR_value,
                      A_row_start, A_column_index, A_value)){
      mexErrMsgTxt("QUERN failure");
   }
   // check options
   int num_options=mxGetM(prhs[2])*mxGetN(prhs[2]);
   const double* options=mxGetPr(prhs[2]);

   // now decide what to do, and do it
   if(strcasecmp("rbfs", funcname)==0){
      if(nlhs!=1) mexErrMsgTxt("one output argument for rbfs: ordering");
      int* column_order=(int*)mxMalloc(n*sizeof(int));
      if(QUERN_get_rbfs_column_ordering(m, n, A_row_start, A_column_index,
                                        column_order))
         mexErrMsgTxt("QUERN failure");
      plhs[0]=create_dense(n, column_order);

   }else if(strcasecmp("reorder", funcname)==0){
      if(nlhs!=1) mexErrMsgTxt("one output argument for reorder: matrix");
      if(num_options!=n) mexErrMsgTxt("options must be ordering vector");
      int* column_order=(int*)mxMalloc(n*sizeof(int));
      for(int i=0; i<n; ++i) column_order[i]=(int)options[i]-1;
      if(QUERN_reorder_columns(m, n, column_order, A_row_start, A_column_index,
                               A_value))
         mexErrMsgTxt("QUERN failure");
      plhs[0]=create_sparse_flipped(m, n, A_row_start, A_column_index, A_value);
   
   }else if(strcasecmp("profile", funcname)==0){
      if(nlhs>1) mexErrMsgTxt("one output argument for profile: ordering");
      int* row_order=(int*)mxMalloc(m*sizeof(int));
      if(QUERN_get_profile_row_ordering(m, n, A_row_start, A_column_index,
                                        row_order))
         mexErrMsgTxt("QUERN failure");
      plhs[0]=create_dense(m, row_order);

   }else if(strcasecmp("r", funcname)==0){
      if(nlhs!=1) mexErrMsgTxt("one output argument: R");
      // get row order from options (or not, or automatic)
      int* row_order=0;
      if(num_options==0){
         // do nothing - use default ordering
      }else if(num_options==1 && options[0]<=0){
         row_order=(int*)mxMalloc(m*sizeof(int));
         if(QUERN_get_profile_row_ordering(m, n, A_row_start, A_column_index,
                                           row_order))
            mexErrMsgTxt("QUERN failure");
      }else{
         if(num_options!=m) mexErrMsgTxt("wrong size for row ordering");
         row_order=(int*)mxMalloc(m*sizeof(int));
         for(int i=0; i<m; ++i) row_order[i]=(int)options[i]-1;
      }
      // get R
      int* R_row_start=0;
      int* R_column_index=0;
      double* R_value=0;
      if(QUERN_compute_qr_without_q(m, n, A_row_start, A_column_index, A_value,
                                    row_order, &R_row_start, &R_column_index,
                                    &R_value))
            mexErrMsgTxt("QUERN failure");
      plhs[0]=create_sparse_flipped(n, n, R_row_start, R_column_index, R_value);
      QUERN_free_result(R_row_start, R_column_index, R_value);
   
   }else if(strcasecmp("qr", funcname)==0){
      if(nlhs!=2) mexErrMsgTxt("two output arguments needed: R, G");
      // get row order from options (or not, or automatic)
      int* row_order=0;
      if(num_options==0){
         // do nothing - use default ordering
      }else if(num_options==1 && options[0]<=0){
         row_order=(int*)mxMalloc(m*sizeof(int));
         if(QUERN_get_profile_row_ordering(m, n, A_row_start, A_column_index,
                                           row_order))
            mexErrMsgTxt("QUERN failure");
      }else{
         if(num_options!=m) mexErrMsgTxt("wrong size for row ordering");
         row_order=(int*)mxMalloc(m*sizeof(int));
         for(int i=0; i<m; ++i) row_order[i]=(int)options[i]-1;
      }
      // get Q and R
      int* Q_row_start=0;
      int* Q_column_index=0;
      double* Q_value=0;
      int* R_row_start=0;
      int* R_column_index=0;
      double* R_value=0;
      if(QUERN_compute_qr(m, n, A_row_start, A_column_index, A_value,
                          row_order, &Q_row_start, &Q_column_index, &Q_value,
                          &R_row_start, &R_column_index, &R_value))
            mexErrMsgTxt("QUERN failure");
      plhs[0]=create_sparse_flipped(n, n, R_row_start, R_column_index, R_value);
      QUERN_free_result(R_row_start, R_column_index, R_value);
      plhs[1]=create_sparse_flipped(m, n, Q_row_start, Q_column_index, Q_value);
      QUERN_free_result(Q_row_start, Q_column_index, Q_value);

   }else if(strcasecmp("ir", funcname)==0){
      if(nlhs!=1) mexErrMsgTxt("one output argument: R");
      // get drop tolerance from options
      double drop_tolerance=0.1;
      if(num_options>=1) drop_tolerance=options[0];
      // use profile row ordering
      int* row_order=(int*)mxMalloc(m*sizeof(int));
      if(QUERN_get_profile_row_ordering(m, n, A_row_start, A_column_index,
               row_order))
         mexErrMsgTxt("QUERN failure");
      // get R
      int* R_row_start=0;
      int* R_column_index=0;
      double* R_value=0;
      if(QUERN_compute_incomplete_qr_without_q(m, n, A_row_start,
                            A_column_index, A_value, row_order, drop_tolerance,
                            &R_row_start, &R_column_index, &R_value))
            mexErrMsgTxt("QUERN failure");
      plhs[0]=create_sparse_flipped(n, n, R_row_start, R_column_index, R_value);
      QUERN_free_result(R_row_start, R_column_index, R_value);

   }else if(strcasecmp("iqr", funcname)==0){
      if(nlhs!=2) mexErrMsgTxt("two output arguments needed: R, G");
      // get drop tolerance from options
      double drop_tolerance=0.1;
      if(num_options>=1) drop_tolerance=options[0];
      // use natural row order - up to user to permute matrix before call
      int* row_order=0;
      // get Q and R
      int* Q_row_start=0;
      int* Q_column_index=0;
      double* Q_value=0;
      int* R_row_start=0;
      int* R_column_index=0;
      double* R_value=0;
      if(QUERN_compute_incomplete_qr(m, n, A_row_start, A_column_index, A_value,
                                     row_order, drop_tolerance,
                                     &Q_row_start, &Q_column_index, &Q_value,
                                     &R_row_start, &R_column_index, &R_value))
         mexErrMsgTxt("QUERN failure");
      plhs[0]=create_sparse_flipped(n, n, R_row_start, R_column_index, R_value);
      QUERN_free_result(R_row_start, R_column_index, R_value);
      plhs[1]=create_sparse_flipped(m, n, Q_row_start, Q_column_index, Q_value);
      QUERN_free_result(Q_row_start, Q_column_index, Q_value);

   }else if(strcasecmp("q*", funcname)==0){
      if(nlhs!=1) mexErrMsgTxt("multiplying by Q requires one output argument");
      if(num_options!=m) mexErrMsgTxt("vector should be full size");
      plhs[0]=mxCreateNumericMatrix(m, 1, mxDOUBLE_CLASS, mxREAL);
      double *x=mxGetPr(plhs[0]);
      std::memcpy(x, options, m*sizeof(double));
      if(QUERN_multiply_with_q(m, A_row_start, A_column_index, A_value, x))
         mexErrMsgTxt("QUERN failure");

   }else if(strcasecmp("q'*", funcname)==0){
      if(nlhs!=1) mexErrMsgTxt("multiplying by Q' requires one output argument");
      if(num_options!=m) mexErrMsgTxt("vector should be full size");
      plhs[0]=mxCreateNumericMatrix(m, 1, mxDOUBLE_CLASS, mxREAL);
      double *x=mxGetPr(plhs[0]);
      std::memcpy(x, options, m*sizeof(double));
      if(QUERN_multiply_with_q_transpose(m, A_row_start, A_column_index,
                                         A_value, x))
         mexErrMsgTxt("QUERN failure");

   }else if(strcasecmp("R\\", funcname)==0){
      if(nlhs!=1) mexErrMsgTxt("solving with R requires one output argument");
      if(num_options!=n) mexErrMsgTxt("vector is wrong length");
      plhs[0]=mxCreateNumericMatrix(n, 1, mxDOUBLE_CLASS, mxREAL);
      double *x=mxGetPr(plhs[0]);
      if(QUERN_solve_with_r(n, A_row_start, A_column_index, A_value,
                            options, x))
         mexErrMsgTxt("QUERN failure");

   }else if(strcasecmp("R'\\", funcname)==0){
      if(nlhs!=1) mexErrMsgTxt("solving with R' requires one output argument");
      if(num_options!=n) mexErrMsgTxt("vector is wrong length");
      plhs[0]=mxCreateNumericMatrix(n, 1, mxDOUBLE_CLASS, mxREAL);
      double *x=mxGetPr(plhs[0]);
      std::memcpy(x, options, n*sizeof(double));
      if(QUERN_solve_with_r_transpose_in_place(n, A_row_start, A_column_index,
                                               A_value, x))
         mexErrMsgTxt("QUERN failure");

   }else{
      mexErrMsgTxt("Unknown function name");
   }

   QUERN_free_result(A_row_start, A_column_index, A_value);
}

