#include <cmath>
#include <iostream>
#include "mat.h"
#include "quern.h"

const double drop_tolerance=0.1;
const int max_iterations=1000;
const double relative_convergence_tolerance=1e-8;

void test_solve(int m,
                int n,
                int* A_row_start,
                int* A_column_index,
                double* A_value)
{
   int* column_order=0;
   int* row_order=0;
   double* b=0;
   double* c=0;
   double* x=0;
   int* R_row_start=0;
   int* R_column_index=0;
   double* R_value=0;
   double tolerance;
   int solved;
   int iterations;
   double residual_norm;
   // do the memory allocation we'll need
   if(!(column_order=(int*)mxMalloc(n*sizeof(int)))) goto finish;
   if(!(row_order=(int*)mxMalloc(m*sizeof(int)))) goto finish;
   if(!(b=(double*)mxMalloc(m*sizeof(double)))) goto finish;
   if(!(c=(double*)mxMalloc(n*sizeof(double)))) goto finish;
   if(!(x=(double*)mxMalloc(n*sizeof(double)))) goto finish;
   // first order columns
   if(QUERN_get_rbfs_column_ordering(m, n, A_row_start, A_column_index, column_order)) goto finish;
   if(QUERN_reorder_columns(m, n, column_order, A_row_start, A_column_index, A_value)) goto finish;
   // then order rows
   if(QUERN_get_profile_row_ordering(m, n, A_row_start, A_column_index, row_order)) goto finish;
   // then do incomplete QR
   if(QUERN_compute_incomplete_qr_without_q(m, n, A_row_start, A_column_index, A_value, row_order, drop_tolerance, &R_row_start, &R_column_index, &R_value)) goto finish;
   // set up a least-squares problem ||Ax-b||
   for(int i=0; i<m; ++i){
      b[i]=std::rand()/(double)RAND_MAX;
   }
   // and form c=A^Tb
   if(QUERN_multiply_transpose(m, n, A_row_start, A_column_index, A_value, b, c)) goto finish;
   tolerance=0;
   for(int i=0; i<n; ++i) tolerance+=c[i]*c[i];
   tolerance=std::sqrt(tolerance)*relative_convergence_tolerance;
   // attempt to solve
   solved=0;
   iterations=0;
   residual_norm=0;
   if(QUERN_solve_with_CGNR(m, n, A_row_start, A_column_index, A_value, c, R_row_start, R_column_index, R_value, max_iterations, tolerance, x, &solved, &iterations, &residual_norm)) goto finish;
   // print out the results
   std::cout<<(solved ? "solved" : "failed to solve")<<std::endl;
   std::cout<<"iterations: "<<iterations<<std::endl;
   std::cout<<"final residual norm: "<<residual_norm<<std::endl;
   // finish up
   finish:
   QUERN_free_result(R_row_start, R_column_index, R_value);
   mxFree(x);
   mxFree(c);
   mxFree(b);
   mxFree(row_order);
   mxFree(column_order);
}

int main(int argc, char** argv)
{
   for(int i=1; i<argc; ++i){
      std::cout<<argv[i]<<" ####################"<<std::endl;
      MATFile *mf=matOpen(argv[i], "r");
      if(!mf){
         std::cout<<" - couldn't open MAT file"<<std::endl;
         continue;
      }
      mxArray* A=matGetVariable(mf, "A");
      if(!A){
         std::cout<<" - couldn't get matrix A in MAT file"<<std::endl;
         matClose(mf);
         continue;
      }
      // check A
      if(mxGetNumberOfDimensions(A)!=2){
         std::cout<<" - A is not two dimensional"<<std::endl;
         matClose(mf);
         continue;
      }
      if(!mxIsSparse(A)){
         std::cout<<" - A is not sparse"<<std::endl;
         matClose(mf);
         continue;
      }
      int m=mxGetM(A), n=mxGetN(A);
      const int* A_column_start=(const int*)mxGetJc(A);
      const int* A_row_index=(const int*)mxGetIr(A);
      const double* A_CSR_value=mxGetPr(A);

      // transpose
      int nnz=A_column_start[n];
      int* A_row_start=(int*)mxMalloc((m+1)*sizeof(int));
      int* A_column_index=(int*)mxMalloc(nnz*sizeof(int));
      double* A_value=(double*)mxMalloc(nnz*sizeof(double));
      int transpose=QUERN_convert_column_format_to_row_format(m, n,
            A_column_start, A_row_index, A_CSR_value,
            A_row_start, A_column_index, A_value);
      if(transpose){
         std::cout<<" - Quern transpose error "<<transpose<<std::endl;
         matClose(mf);
         continue;
      }

      // finish with file
      matClose(mf);

      // check out matrix
      test_solve(m, n, A_row_start, A_column_index, A_value);
   }
   return 0;
}
